from tdw.FBOutput import Vector3, Quaternion, PassMask, Color, MessageType, MachineType, SimpleTransform
from tdw.FBOutput import Environments as Envs
from tdw.FBOutput import Transforms as Trans
from tdw.FBOutput import Rigidbodies as Rigis
from tdw.FBOutput import Bounds as Bouns
from tdw.FBOutput import Images as Imags
from tdw.FBOutput import AvatarKinematic as AvKi
from tdw.FBOutput import AvatarNonKinematic as AvNoKi
from tdw.FBOutput import AvatarSimpleBody as AvSi
from tdw.FBOutput import AvatarStickyMitten as AvSM
from tdw.FBOutput import SegmentationColors as Segs
from tdw.FBOutput import AvatarSegmentationColor as AvSC
from tdw.FBOutput import AvatarStickyMittenSegmentationColors as AvSMSC
from tdw.FBOutput import IsOnNavMesh as IsNM
from tdw.FBOutput import IdPassGrayscale as IdGS
from tdw.FBOutput import Collision as Col
from tdw.FBOutput import ImageSensors as ImSe
from tdw.FBOutput import CameraMatrices as CaMa
from tdw.FBOutput import IdPassSegmentationColors as IdSC
from tdw.FBOutput import ArrivedAtNavMeshDestination as Arri
from tdw.FBOutput import FlexParticles as Flex
from tdw.FBOutput import VRRig as VR
from tdw.FBOutput import LogMessage as Log
from tdw.FBOutput import CompositeObjects as Comp
from tdw.FBOutput import Meshes as Me
from tdw.FBOutput import Substructure as Sub
from tdw.FBOutput import Version as Ver
from tdw.FBOutput import EnvironmentCollision as EnvCol
from tdw.FBOutput import Volumes as Vol
from tdw.FBOutput import AudioSources as Audi
from tdw.FBOutput import AvatarChildrenNames as AvCN
from tdw.FBOutput import Raycast as Ray
import numpy as np
from typing import Tuple, Optional


class OutputDataUndefinedError(Exception):
    pass


class OutputData(object):
    def __init__(self, b):
        self.bytes = bytearray(b)
        self.data = self.get_data()

    def get_data(self):
        raise OutputDataUndefinedError("Undefined!")

    @staticmethod
    def get_data_type_id(b: bytes) -> str:
        """
        Returns the ID of the serialized object.
        :param b: A byte array.
        """

        return b[4:8].decode('utf-8')

    @staticmethod
    def _get_vector3(constructor) -> Tuple[float, float, float]:
        """
        Returns x, y, and z values of a Vector3, given a constructor.

        :param constructor: A constructor that accepts 1 parameter of type Vector3.
        """

        return OutputData._get_xyz(constructor(Vector3.Vector3()))

    @staticmethod
    def _get_xyz(vector3: Vector3) -> Tuple[float, float, float]:
        """
        returns the x, y, and z values of a Vector3, given the Vector3 object.

        :param vector3: The Vector3 object.
        """

        return vector3.X(), vector3.Y(), vector3.Z()

    @staticmethod
    def _get_quaternion(constructor) -> Tuple[float, float, float, float]:
        """
        Returns x, y, z, and w values of a Quaternion, given a constructor.

        :param constructor: A constructor that accepts 1 parameter of type Quaternion.
        """

        return OutputData._get_xyzw(constructor(Quaternion.Quaternion()))

    @staticmethod
    def _get_xyzw(quaternion: Quaternion) -> Tuple[float, float, float, float]:
        """
        returns the x, y, and z values of a Quaternion, given the Quaternion object.

        :param quaternion: The Quaternion object.
        """

        return quaternion.X(), quaternion.Y(), quaternion.Z(), quaternion.W()

    @staticmethod
    def _get_color(constructor) -> Tuple[float, float, float]:
        """
        Returns the r, g, and b values of a Color, given a constructor.

        :param constructor: A constructor that accepts 1 parameter of type Color.
        """

        return OutputData._get_rgb(constructor(Color.Color()))

    @staticmethod
    def _get_rgb(color: Color) -> Tuple[float, float, float]:
        """
        returns the r, g, and b values of a Color, given the Color object.

        :param color: The Color object.
        """
        return color.R(), color.G(), color.B()


class Environments(OutputData):
    def get_data(self) -> Envs.Environments:
        return Envs.Environments.GetRootAsEnvironments(self.bytes, 0)

    def get_center(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Envs(index).Center)

    def get_bounds(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Envs(index).Bounds)

    def get_id(self, index: int) -> int:
        return self.data.Envs(index).Id()

    def get_num(self) -> int:
        return self.data.EnvsLength()


class Transforms(OutputData):
    def get_data(self) -> Trans.Transforms:
        return Trans.Transforms.GetRootAsTransforms(self.bytes, 0)

    def get_num(self) -> int:
        return self.data.ObjectsLength()

    def get_id(self, index: int) -> int:
        return self.data.Objects(index).Id()

    def get_position(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Objects(index).Position)

    def get_forward(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Objects(index).Forward)

    def get_rotation(self, index: int) -> Tuple[float, float, float, float]:
        return OutputData._get_quaternion(self.data.Objects(index).Rotation)


class Rigidbodies(OutputData):
    def get_data(self) -> Rigis.Rigidbodies:
        return Rigis.Rigidbodies.GetRootAsRigidbodies(self.bytes, 0)

    def get_num(self) -> int:
        return self.data.ObjectsLength()

    def get_id(self, index: int) -> int:
        return self.data.Objects(index).Id()

    def get_velocity(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Objects(index).Velocity)

    def get_angular_velocity(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Objects(index).AngularVelocity)

    def get_mass(self, index: int) -> float:
        return self.data.Objects(index).Mass()

    def get_sleeping(self, index: int) -> bool:
        return self.data.Objects(index).Sleeping()


class Bounds(OutputData):
    def get_data(self) -> Bouns.Bounds:
        return Bouns.Bounds.GetRootAsBounds(self.bytes, 0)

    def get_num(self) -> int:
        return self.data.ObjectsLength()

    def get_id(self, index: int) -> int:
        return self.data.Objects(index).Id()

    def get_front(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Objects(index).Front)

    def get_back(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Objects(index).Back)

    def get_left(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Objects(index).Left)

    def get_right(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Objects(index).Right)

    def get_top(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Objects(index).Top)

    def get_bottom(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Objects(index).Bottom)

    def get_center(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Objects(index).Center)


class Images(OutputData):
    PASS_MASKS = {PassMask.PassMask._img: "_img",
                  PassMask.PassMask._id: "_id",
                  PassMask.PassMask._category: "_category",
                  PassMask.PassMask._mask: "_mask",
                  PassMask.PassMask._depth: "_depth",
                  PassMask.PassMask._normals: "_normals",
                  PassMask.PassMask._flow: "_flow",
                  PassMask.PassMask._depth_simple: "_depth_simple"
                  }

    def get_data(self) -> Imags.Images:
        return Imags.Images.GetRootAsImages(self.bytes, 0)

    def get_avatar_id(self) -> str:
        return self.data.AvatarId().decode('utf-8')

    def get_sensor_name(self) -> str:
        return self.data.SensorName().decode('utf-8')

    def get_num_passes(self) -> int:
        return self.data.PassesLength()

    def get_pass_mask(self, index: int) -> str:
        return Images.PASS_MASKS[self.data.Passes(index).PassMask()]

    def get_image(self, index: int) -> np.array:
        return self.data.Passes(index).ImageAsNumpy()

    def get_extension(self, index: int) -> str:
        return "png" if self.data.Passes(index).Extension() == 1 else "jpg"


class AvatarKinematic(OutputData):
    def get_data(self) -> AvKi.AvatarKinematic:
        return AvKi.AvatarKinematic.GetRootAsAvatarKinematic(self.bytes, 0)

    def get_avatar_id(self) -> str:
        return self.data.Id().decode('utf-8')

    def get_position(self) -> Tuple[float, float, float]:
        return OutputData._get_xyz(self.data.Position())

    def get_rotation(self) -> Tuple[float, float, float, float]:
        return OutputData._get_xyzw(self.data.Rotation())

    def get_forward(self) -> Tuple[float, float, float]:
        return OutputData._get_xyz(self.data.Forward())


class AvatarNonKinematic(AvatarKinematic):
    def get_data(self) -> AvNoKi.AvatarNonKinematic:
        return AvNoKi.AvatarNonKinematic.GetRootAsAvatarNonKinematic(self.bytes, 0)

    def get_velocity(self) -> Tuple[float, float, float]:
        return OutputData._get_xyz(self.data.Velocity())

    def get_angular_velocity(self) -> Tuple[float, float, float]:
        return OutputData._get_xyz(self.data.AngularVelocity())

    def get_mass(self) -> float:
        return self.data.Mass()

    def get_sleeping(self) -> bool:
        return self.data.Sleeping()


class AvatarSimpleBody(AvatarNonKinematic):
    def get_data(self) -> AvSi.AvatarSimpleBody:
        return AvSi.AvatarSimpleBody.GetRootAsAvatarSimpleBody(self.bytes, 0)

    def get_visible_body(self) -> str:
        return self.data.VisibleBody().decode('utf-8')


class AvatarStickyMitten(AvatarNonKinematic):
    def get_data(self) -> AvSM.AvatarStickyMitten:
        return AvSM.AvatarStickyMitten.GetRootAsAvatarStickyMitten(self.bytes, 0)

    def get_num_body_parts(self) -> int:
        return self.data.BodyPartsLength()

    def get_num_rigidbody_parts(self) -> int:
        return self.data.RigidbodyPartsLength()

    def get_body_part_position(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.BodyParts(index).Position)

    def get_body_part_rotation(self, index: int) -> Tuple[float, float, float, float]:
        return OutputData._get_quaternion(self.data.BodyParts(index).Rotation)

    def get_body_part_forward(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.BodyParts(index).Forward)

    def get_body_part_id(self, index: int) -> int:
        return self.data.BodyParts(index).Id()

    def get_rigidbody_part_velocity(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.RigidbodyParts(index).Velocity)

    def get_rigidbody_part_angular_velocity(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.RigidbodyParts(index).AngularVelocity)

    def get_rigidbody_part_mass(self, index: int) -> float:
        return self.data.RigidbodyParts(index).Mass()

    def get_rigidbody_part_sleeping(self, index: int) -> bool:
        return self.data.RigidbodyParts(index).Sleeping()

    def get_rigidbody_part_id(self, index: int) -> int:
        return self.data.RigidbodyParts(index).Id()

    def get_held_left(self) -> np.array:
        return self.data.HeldLeftAsNumpy()

    def get_held_right(self) -> np.array:
        return self.data.HeldRightAsNumpy()

    def get_angles_left(self) -> np.array:
        return self.data.AnglesLeftAsNumpy()

    def get_angles_right(self) -> np.array:
        return self.data.AnglesRightAsNumpy()

    def get_mitten_center_left_position(self) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.MittenCenterLeft().Position)

    def get_mitten_center_left_forward(self) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.MittenCenterLeft().Forward)

    def get_mitten_center_left_rotation(self) -> Tuple[float, float, float, float]:
        return OutputData._get_quaternion(self.data.MittenCenterLeft().Rotation)

    def get_mitten_center_right_position(self) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.MittenCenterRight().Position)

    def get_mitten_center_right_forward(self) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.MittenCenterRight().Forward)

    def get_mitten_center_right_rotation(self, index: int) -> Tuple[float, float, float, float]:
        return OutputData._get_quaternion(self.data.MittenCenterRight(index).Rotation)


class SegmentationColors(OutputData):
    def get_data(self) -> Segs.SegmentationColors:
        return Segs.SegmentationColors.GetRootAsSegmentationColors(self.bytes, 0)

    def get_num(self) -> int:
        return self.data.ObjectsLength()

    def get_object_id(self, index: int) -> int:
        return self.data.Objects(index).Id()

    def get_object_color(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_rgb(self.data.Objects(index).SegmentationColor())

    def get_object_name(self, index: int) -> str:
        return self.data.Objects(index).Name().decode('utf-8')


class AvatarSegmentationColor(OutputData):
    def get_data(self) -> AvSC.AvatarSegmentationColor:
        return AvSC.AvatarSegmentationColor.GetRootAsAvatarSegmentationColor(self.bytes, 0)

    def get_id(self) -> str:
        return self.data.Id().decode('utf-8')

    def get_segmentation_color(self) -> Tuple[float, float, float]:
        return OutputData._get_rgb(self.data.SegmentationColor())


class AvatarStickyMittenSegmentationColors(OutputData):
    def get_data(self) -> AvSMSC.AvatarStickyMittenSegmentationColors:
        return AvSMSC.AvatarStickyMittenSegmentationColors.GetRootAsAvatarStickyMittenSegmentationColors(self.bytes, 0)

    def get_id(self) -> str:
        return self.data.Id().decode('utf-8')

    def get_num_body_parts(self) -> int:
        return self.data.BodyPartsLength()

    def get_body_part_id(self, index: int) -> int:
        return self.data.BodyParts(index).Id()

    def get_body_part_segmentation_color(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_rgb(self.data.BodyParts(index).SegmentationColor())

    def get_body_part_name(self, index: int) -> str:
        return self.data.BodyParts(index).Name().decode('utf-8')


class IsOnNavMesh(OutputData):
    def get_data(self) -> IsNM.IsOnNavMesh:
        return IsNM.IsOnNavMesh.GetRootAsIsOnNavMesh(self.bytes, 0)

    def get_position(self) -> Tuple[float, float, float]:
        return OutputData._get_xyz(self.data.Position())

    def get_is_on(self) -> bool:
        return self.data.IsOn()


class IdPassGrayscale(OutputData):
    def get_data(self) -> IdGS.IdPassGrayscale:
        return IdGS.IdPassGrayscale.GetRootAsIdPassGrayscale(self.bytes, 0)

    def get_avatar_id(self) -> str:
        return self.data.AvatarId().decode('utf-8')

    def get_sensor_name(self) -> str:
        return self.data.SensorName().decode('utf-8')

    def get_grayscale(self) -> float:
        return self.data.Grayscale()


class Collision(OutputData):
    def get_data(self) -> Col.Collision:
        return Col.Collision.GetRootAsCollision(self.bytes, 0)

    def get_collider_id(self) -> int:
        return self.data.ColliderId()

    def get_collidee_id(self) -> int:
        return self.data.CollideeId()

    def get_relative_velocity(self) -> Tuple[float, float, float]:
        return OutputData._get_xyz(self.data.RelativeVelocity())

    def get_state(self) -> str:
        state = self.data.State()
        if state == 1:
            return "enter"
        elif state == 2:
            return "stay"
        else:
            return "exit"

    def get_num_contacts(self) -> int:
        return self.data.ContactsLength()

    def get_contact_normal(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Contacts(index).Normal)

    def get_contact_point(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Contacts(index).Point)


class ImageSensors(OutputData):
    def get_data(self) -> ImSe.ImageSensors:
        return ImSe.ImageSensors.GetRootAsImageSensors(self.bytes, 0)

    def get_avatar_id(self) -> str:
        return self.data.AvatarId().decode('utf-8')

    def get_num_sensors(self) -> int:
        return self.data.SensorsLength()

    def get_sensor_name(self, index: int) -> str:
        return self.data.Sensors(index).Name().decode('utf-8')

    def get_sensor_on(self, index: int) -> bool:
        return self.data.Sensors(index).IsOn()

    def get_sensor_rotation(self, index: int) -> Tuple[float, float, float, float]:
        return OutputData._get_xyzw(self.data.Sensors(index).Rotation())


class CameraMatrices(OutputData):
    def get_data(self) -> CaMa.CameraMatrices:
        return CaMa.CameraMatrices.GetRootAsCameraMatrices(self.bytes, 0)

    def get_avatar_id(self) -> str:
        return self.data.AvatarId().decode('utf-8')

    def get_sensor_name(self) -> str:
        return self.data.SensorName().decode('utf-8')

    def get_projection_matrix(self) -> np.array:
        return self.data.ProjectionMatrixAsNumpy()

    def get_camera_matrix(self) -> np.array:
        return self.data.CameraMatrixAsNumpy()


class IdPassSegmentationColors(OutputData):
    def get_data(self) -> IdSC.IdPassSegmentationColors:
        return IdSC.IdPassSegmentationColors.GetRootAsIdPassSegmentationColors(self.bytes, 0)

    def get_avatar_id(self) -> str:
        return self.data.AvatarId().decode('utf-8')

    def get_sensor_name(self) -> str:
        return self.data.SensorName().decode('utf-8')

    def get_num_segmentation_colors(self) -> int:
        return self.data.SegmentationColorsLength()

    def get_segmentation_color(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_rgb(self.data.SegmentationColors(index))


class ArrivedAtNavMeshDestination(OutputData):
    def get_data(self) -> Arri.ArrivedAtNavMeshDestination:
        return Arri.ArrivedAtNavMeshDestination.GetRootAsArrivedAtNavMeshDestination(self.bytes, 0)

    def get_avatar_id(self) -> str:
        return self.data.AvatarId().decode('utf-8')


class FlexParticles(OutputData):
    def get_data(self) -> Flex.FlexParticles:
        return Flex.FlexParticles.GetRootAsFlexParticles(self.bytes, 0)

    def get_num_objects(self) -> int:
        return self.data.ObjectsLength()

    def get_particles(self, index: int) -> np.array:
        return self.data.Objects(index).ParticlesAsNumpy().view(dtype=np.float32).reshape(-1, 4)

    def get_velocities(self, index: int) -> np.array:
        return self.data.Objects(index).VelocitiesAsNumpy().view(dtype=np.float32).reshape(-1, 3)

    def get_id(self, index: int) -> int:
        return self.data.Objects(index).Id()


class VRRig(OutputData):
    def get_data(self) -> VR.VRRig:
        return VR.VRRig.GetRootAsVRRig(self.bytes, 0)

    def get_position(self) -> Tuple[float, float, float]:
        return OutputData._get_xyz(self.data.Position())

    def get_rotation(self) -> Tuple[float, float, float, float]:
        return OutputData._get_xyzw(self.data.Rotation())

    def get_forward(self) -> Tuple[float, float, float]:
        return OutputData._get_xyz(self.data.Forward())

    def _get_simple_transform(self, t: int) -> SimpleTransform:
        if t == 0:
            return self.data.LeftHand()
        elif t == 1:
            return self.data.RightHand()
        elif t == 2:
            return self.data.Head()
        else:
            raise Exception("Not defined: " + str(t))

    def _get_hand_position(self, is_left: bool) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self._get_simple_transform(0 if is_left else 1).Position)

    def _get_hand_rotation(self, is_left: bool) -> Tuple[float, float, float, float]:
        return OutputData._get_quaternion(self._get_simple_transform(0 if is_left else 1).Rotation)

    def _get_hand_forward(self, is_left: bool) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self._get_simple_transform(0 if is_left else 1).Forward)

    def get_left_hand_position(self) -> Tuple[float, float, float]:
        return self._get_hand_position(True)

    def get_left_hand_rotation(self) -> Tuple[float, float, float, float]:
        return self._get_hand_rotation(True)

    def get_left_hand_forward(self) -> Tuple[float, float, float]:
        return self._get_hand_forward(True)

    def get_right_hand_position(self) -> Tuple[float, float, float]:
        return self._get_hand_position(False)

    def get_right_hand_rotation(self) -> Tuple[float, float, float, float]:
        return self._get_hand_rotation(False)

    def get_right_hand_forward(self) -> Tuple[float, float, float]:
        return self._get_hand_forward(False)

    def get_head_position(self) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self._get_simple_transform(2).Position)

    def get_head_rotation(self) -> Tuple[float, float, float, float]:
        return OutputData._get_quaternion(self._get_simple_transform(2).Rotation)

    def get_head_forward(self) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self._get_simple_transform(2).Forward)


class LogMessage(OutputData):
    LOG_TYPES = {MessageType.MessageType.error: "error",
                 MessageType.MessageType.warning: "warning",
                 MessageType.MessageType.message: "message",
                 }

    def get_data(self) -> Log.LogMessage:
        return Log.LogMessage.GetRootAsLogMessage(self.bytes, 0)

    def get_message(self) -> str:
        return self.data.Message().decode('utf-8')

    def get_message_type(self) -> str:
        return LogMessage.LOG_TYPES[self.data.MessageType()]

    def get_object_type(self) -> str:
        return self.data.ObjectType().decode('utf-8')


class CompositeObjects(OutputData):
    MACHINE_TYPES = {MachineType.MachineType.light: "light",
                     MachineType.MachineType.motor: "motor",
                     MachineType.MachineType.hinge: "hinge",
                     MachineType.MachineType.spring: "spring",
                     MachineType.MachineType.none: "none"}

    def get_data(self) -> Comp.CompositeObjects:
        return Comp.CompositeObjects.GetRootAsCompositeObjects(self.bytes, 0)

    def get_num(self) -> int:
        return self.data.ObjectsLength()

    def get_object_id(self, index: int) -> int:
        return self.data.Objects(index).Id()

    def get_num_sub_objects(self, index: int) -> int:
        return self.data.Objects(index).SubObjectsLength()

    def get_sub_object_id(self, index: int, sub_object_index: int) -> int:
        return self.data.Objects(index).SubObjects(sub_object_index).Id()

    def get_sub_object_machine_type(self, index: int, sub_object_index: int) -> str:
        return CompositeObjects.MACHINE_TYPES[self.data.Objects(index).SubObjects(sub_object_index).MachineType()]


class Meshes(OutputData):
    def get_data(self) -> Me.Meshes:
        return Me.Meshes.GetRootAsMeshes(self.bytes, 0)

    def get_object_id(self, index: int) -> int:
        return self.data.Objects(index).Id()

    def get_num(self) -> int:
        return self.data.ObjectsLength()

    def get_vertices(self, index: int) -> np.array:
        return self.data.Objects(index).VerticesAsNumpy().view(dtype=np.float32).reshape(-1, 3)

    def get_triangles(self, index: int) -> np.array:
        return self.data.Objects(index).TrianglesAsNumpy().view(dtype=np.int32).reshape(-1, 3)


class Substructure(OutputData):
    def get_data(self) -> Sub.Substructure:
        return Sub.Substructure.GetRootAsSubstructure(self.bytes, 0)

    def get_num_sub_objects(self) -> int:
        return self.data.SubObjectsLength()

    def get_sub_object_name(self, index: int) -> str:
        return self.data.SubObjects(index).Name().decode('utf-8')

    def get_num_sub_object_materials(self, index: int) -> int:
        return self.data.SubObjects(index).MaterialsLength()

    def get_sub_object_material(self, index: int, material_index: int) -> str:
        return self.data.SubObjects(index).Materials(material_index).decode('utf-8')


class Version(OutputData):
    def get_data(self) -> Ver.Version:
        return Ver.Version.GetRootAsVersion(self.bytes, 0)

    def get_unity_version(self) -> str:
        return self.data.Unity().decode('utf-8')

    def get_tdw_version(self) -> str:
        return self.data.Tdw().decode('utf-8')

    def get_standalone(self) -> bool:
        return self.data.Standalone()


class EnvironmentCollision(OutputData):
    def get_data(self) -> EnvCol.EnvironmentCollision:
        return EnvCol.EnvironmentCollision.GetRootAsEnvironmentCollision(self.bytes, 0)

    def get_object_id(self) -> int:
        return self.data.ObjectId()

    def get_state(self) -> str:
        state = self.data.State()
        if state == 1:
            return "enter"
        elif state == 2:
            return "stay"
        else:
            return "exit"

    def get_num_contacts(self) -> int:
        return self.data.ContactsLength()

    def get_contact_normal(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Contacts(index).Normal)

    def get_contact_point(self, index: int) -> Tuple[float, float, float]:
        return OutputData._get_vector3(self.data.Contacts(index).Point)


class Volumes(OutputData):
    def get_data(self) -> Vol.Volumes:
        return Vol.Volumes.GetRootAsVolumes(self.bytes, 0)

    def get_num(self) -> int:
        return self.data.ObjectsLength()

    def get_object_id(self, index: int) -> int:
        return self.data.Objects(index).Id()

    def get_volume(self, index: int) -> float:
        return self.data.Objects(index).Volume()


class AudioSources(OutputData):
    def get_data(self) -> Audi.AudioSources:
        return Audi.AudioSources.GetRootAsAudioSources(self.bytes, 0)

    def get_num(self) -> int:
        return self.data.ObjectsLength()

    def get_object_id(self, index: int) -> int:
        return self.data.Objects(index).Id()

    def get_is_playing(self, index: int) -> bool:
        return self.data.Objects(index).IsPlaying()


class AvatarChildrenNames(OutputData):
    def get_data(self) -> AvCN.AvatarChildrenNames:
        return AvCN.AvatarChildrenNames.GetRootAsAvatarChildrenNames(self.bytes, 0)

    def get_avatar_id(self) -> str:
        return self.data.Id().decode('utf-8')

    def get_num_children(self) -> int:
        return self.data.ChildIdsLength()

    def get_child_name(self, index: int) -> str:
        return self.data.ChildNames(index).decode('utf-8')

    def get_child_id(self, index: int) -> int:
        return self.data.ChildIds(index)


class Raycast(OutputData):
    def get_data(self) -> Ray.Raycast:
        return Ray.Raycast.GetRootAsRaycast(self.bytes, 0)

    def get_raycast_id(self) -> int:
        return self.data.RaycastId()

    def get_hit(self) -> bool:
        return self.data.Hit()

    def get_object_id(self) -> Optional[int]:
        return self.data.ObjectId()

    def get_normal(self) -> Tuple[float, float, float]:
        return OutputData._get_xyz(self.data.Normal())

    def get_point(self) -> Tuple[float, float, float]:
        return OutputData._get_xyz(self.data.Point())



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class SubObject(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsSubObject(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = SubObject()
        x.Init(buf, n + offset)
        return x

    # SubObject
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # SubObject
    def Id(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # SubObject
    def MachineType(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 1

def SubObjectStart(builder): builder.StartObject(2)
def SubObjectAddId(builder, id): builder.PrependInt32Slot(0, id, 0)
def SubObjectAddMachineType(builder, machineType): builder.PrependUint8Slot(1, machineType, 1)
def SubObjectEnd(builder): return builder.EndObject()



from typing import List
from subprocess import Popen, PIPE
import re
from pkg_resources import get_distribution


class PyPi:
    """
    Compare the version of the installed tdw Python module to the PyPi version.
    """

    @staticmethod
    def strip_post_release(v: str) -> str:
        """
        If the version number has a post-release suffix (a fourth number), strip it.

        :param v: The version number.

        :return: The version, stripped of the post-release suffix.
        """

        if len(v.split(".")) > 3:
            return '.'.join(v.split('.')[:3])
        else:
            return v

    @staticmethod
    def get_major_release(v: str) -> str:
        """
        :param v: The version number.

        :return: The major release number (example: in 1.7.0, the major release is 7).
        """

        return v.split(".")[1].strip()

    @staticmethod
    def _get_pypi_releases() -> List[str]:
        """
        :return: A list of all available PyPi releases.
        """

        # Get an error from  PyPi which will list all available versions.
        p = Popen(["pip3", "install", "tdw=="], stderr=PIPE, stdout=PIPE)
        p.wait()
        stdout, stderr = p.communicate()
        # From the list of available versions, get the last one (the most recent).
        versions = re.search(r"\(from versions: (.*)\)", stderr.decode("utf-8")).group(1).split(",")
        return [v.strip() for v in versions]

    @staticmethod
    def get_pypi_version(truncate: bool = False) -> str:
        """
        :param truncate: If true, remove the post-release number (the fourth number) if there is one.

        :return: The newest available tdw release on PyPi.
        """

        # From the list of available versions, get the last one (the most recent).
        v = PyPi._get_pypi_releases()[-1]

        # Strip the post-release suffix.
        if truncate:
            return PyPi.strip_post_release(v)
        else:
            return v

    @staticmethod
    def get_installed_tdw_version(truncate: bool = False) -> str:
        """
        :param truncate: If true, remove the post-release number (the fourth number) if there is one.

        :return: The version of the tdw Python module installed on this machine.
        """

        v = get_distribution("tdw").version

        # Strip the post-release suffix.
        if truncate:
            return PyPi.strip_post_release(v)
        else:
            return v

    @staticmethod
    def get_latest_post_release(v: str) -> str:
        """
        :param v: A three-part version string, e.g. 1.6.1

        :return: The most up-to-date version or post-release of the tdw module on PyPi with `v`, e.g. 1.6.1.10
        """

        releases = PyPi._get_pypi_releases()
        releases = sorted([r for r in releases if r.startswith(v)])
        if len(releases) == 0:
            return ""
        return releases[-1]

    @staticmethod
    def get_latest_minor_release(v: str) -> str:
        """
        :param v: The version number.

        :return: The most up-to-date version in this major release. (Example: if v == 1.5.0, this returns 1.5.5)
        """

        version = PyPi.strip_post_release(v)
        releases = PyPi._get_pypi_releases()
        releases = sorted([r for r in releases if r.startswith("1." + PyPi.get_major_release(version))])
        if len(releases) == 0:
            return ""
        return releases[-1]



from typing import Dict

# Convert platform.system() to S3 URL infixes.
SYSTEM_TO_S3: Dict[str, str] = {"Windows": "windows",
                                "Darwin": "osx",
                                "Linux": "linux"}
# Convert S3 URL infixes to Unity build targets.
S3_TO_UNITY: Dict[str, str] = {"windows": "StandaloneWindows64",
                               "osx": "StandaloneOSX",
                               "linux": "StandaloneLinux64"}
# Convert platform.system() to Unity build targets.
SYSTEM_TO_UNITY: Dict[str, str] = {"Windows": "StandaloneWindows64",
                                   "Darwin": "StandaloneOSX",
                                   "Linux": "StandaloneLinux64"}
# Convert Unity build targets to platform.system()
UNITY_TO_SYSTEM: Dict[str, str] = {"StandaloneWindows64": "Windows",
                                   "StandaloneOSX": "Darwin",
                                   "StandaloneLinux64": "Linux"}

# Convert the system platform to the build's executable extension.
SYSTEM_TO_EXECUTABLE: Dict[str, str] = {"Windows": ".exe",
                                        "Darwin": ".app",
                                        "Linux": ".x86_64"}

# Convert the system platform to the name of the zipped release file on GitHub.
SYSTEM_TO_RELEASE: Dict[str, str] = {"Windows": "TDW_Windows",
                                     "Darwin": "TDW_OSX",
                                     "Linux": "TDW_Linux"}



import json
from typing import List, Dict, TypeVar, Union, Generic, Optional, Tuple
import pkg_resources
from pathlib import Path
import platform
from secrets import token_hex


class _Record:
    """
    Abstract class for a metadata record.
    """

    _PLATFORM = platform.system()

    def __init__(self, data: Optional[dict] = None):
        """
        :param data: JSON data for the record. If None, the record will initialize with default values.
        """

        if data is None:
            self.name: str = ""
            self.urls: Dict[str, str] = {"Windows": "", "Darwin": "", "Linux": ""}
        else:
            self.name = data["name"]
            self.urls: Dict[str, str] = data["urls"]

    def get_url(self) -> str:
        """
        Returns the URL of the asset bundle for this platform. This is a wrapper for record.urls.
        """

        return self.urls[_Record._PLATFORM]

    def get_serializable(self) -> dict:
        """
        Returns the serializable dictionary of this record.
        """

        return self.__dict__


class ModelRecord(_Record):
    """
    A record of a model asset bundle.
    """

    def __init__(self, data: Optional[dict] = None):
        super().__init__(data)

        if data is None:
            self.wnid: str = ""
            self.wcategory: str = ""
            self.scale_factor: float = 1
            self.do_not_use: bool = False
            self.do_not_use_reason: str = ""
            self.flex: bool = False
            self.substructure: List[dict] = []
            self.bounds: Dict[str, Dict[str, float]] = {"back": {"x": 0, "y": 0, "z": 0},
                                                        "bottom": {"x": 0, "y": 0, "z": 0},
                                                        "center": {"x": 0, "y": 0, "z": 0},
                                                        "front": {"x": 0, "y": 0, "z": 0},
                                                        "left": {"x": 0, "y": 0, "z": 0},
                                                        "right": {"x": 0, "y": 0, "z": 0},
                                                        "top": {"x": 0, "y": 0, "z": 0}}
            self.canonical_rotation: Dict[str, float] = {"x": 0, "y": 0, "z": 0}
            self.physics_quality: float = -1
            self.asset_bundle_sizes: Dict[str, int] = {"Windows": -1, "Darwin": -1, "Linux": -1}
            self.composite_object = False
        else:
            self.wnid: str = data["wnid"]
            self.wcategory: str = data["wcategory"]
            self.scale_factor: float = data["scale_factor"]
            self.do_not_use: bool = data["do_not_use"]
            self.do_not_use_reason: str = data["do_not_use_reason"]
            self.flex: bool = data["flex"]
            self.substructure: List[dict] = data["substructure"]
            self.bounds: Dict[str, Dict[str, float]] = data["bounds"]
            self.canonical_rotation: Dict[str, float] = data["canonical_rotation"]
            self.physics_quality: float = data["physics_quality"]
            self.asset_bundle_sizes: Dict[str, int] = data["asset_bundle_sizes"]
            self.composite_object: bool = data["composite_object"]


class MaterialRecord(_Record):
    """
    A record of a visual material asset bundle.
    """

    def __init__(self, data: Optional[dict] = None):
        super().__init__(data)

        if data is None:
            self.type: str = "Ceramic"
        else:
            self.type: str = data["type"]


class SceneRecord(_Record):
    """
    A record of a scene asset bundle.
    """

    def __init__(self, data: Optional[dict] = None):
        super().__init__(data)

        if data is None:
            self.description: str = ""
            self.hdri: bool = False
            self.location: str = ""
        else:
            self.description: str = data["description"]
            self.hdri: bool = data["hdri"]
            self.location: str = data["location"]


class HDRISkyboxRecord(_Record):
    """
    A record of an HDRI skybox asset bundle.
    """

    def __init__(self, data: Optional[dict] = None):
        super().__init__(data)

        if data is None:
            self.color_temperature: float = 0
            self.sun_elevation: float = 0
            self.sun_initial_angle: float = 0
            self.sun_intensity: float = 0
            self.initial_skybox_rotation: float = 0
            self.exposure: float = 0
            self.location: str = ""
        else:
            self.color_temperature: float = data["color_temperature"]
            self.sun_elevation: float = data["sun_elevation"]
            self.sun_initial_angle: float = data["sun_initial_angle"]
            self.sun_intensity: float = data["sun_intensity"]
            self.initial_skybox_rotation: float = data["initial_skybox_rotation"]
            self.exposure: float = data["exposure"]
            self.location: str = data["location"]


class HumanoidAnimationRecord(_Record):
    """
    A record for a humanoid animation asset bundle.
    """

    def __init__(self, data: Optional[dict] = None):
        super().__init__(data)

        if data is None:
            self.duration: float = 0
            self.loop: bool = False
            self.framerate: int = 0
        else:
            self.duration: float = data["duration"]
            self.loop: bool = data["loop"]
            self.framerate: int = data["framerate"]

    def get_num_frames(self) -> int:
        """
        Returns the number of frames, given the duration and framerate.
        """

        return int(self.duration * self.framerate)


class HumanoidRecord(_Record):
    """
    A record for a humanoid asset bundle.
    """

    def __init__(self, data: Optional[dict] = None):
        super().__init__(data)


T = TypeVar("T", bound=_Record)


class _Librarian(Generic[T]):
    """
    Base abstract class for a metadata librarian.
    """

    def __init__(self, library: str = ""):
        """
        :param library: The absolute path to the library .json file. If empty, a default path in the tdw module will be used.
        """

        if library == "":
            self.library = pkg_resources.resource_filename(__name__, "metadata_libraries/" + self.get_default_library())
        else:
            module_path = pkg_resources.resource_filename(__name__, "metadata_libraries/" + library)
            if Path(module_path).exists():
                self.library = module_path
            else:
                self.library = library

        with open(self.library, "rt") as f:
            self.data = json.load(f)

        self.description = self.data["description"]

        self.records: List[T] = []
        for key in self.data["records"]:
            record = self._generate_record(self.data["records"][key])
            temp_urls = dict()
            # De-localize URLs
            for p in record.urls:
                # Set an absolute path.
                absolute = False
                for prefix in ["file:///", "http://", "https://"]:
                    if record.urls[p].startswith(prefix):
                        temp_urls[p] = record.urls[p]
                        absolute = True
                # De-localize a local path.
                if not absolute:
                    temp_urls[p] = f"file:///{str(Path(self.library).parent.joinpath(record.urls[p]).resolve())}"
                temp_urls[p] = temp_urls[p].replace("\\", "/")
            record.urls = temp_urls
            self.records.append(record)

    def get_default_library(self) -> str:
        """
        Returns the default library path (which is always the first in the list of `get_library_filenames()`)
        """

        return self.get_library_filenames()[0]

    @staticmethod
    def create_library(description: str, path: str) -> None:
        """
        Create a new library JSON file.

        :param path: The absolute filepath to the .json records database file.
        :param description: A brief description of the library.
        """

        path = Path(path)
        data = {"description": description,
                "records": {}}
        path.write_text(json.dumps(data), encoding="utf-8")
        print(f"Created new library: {path}")

    @staticmethod
    def get_library_filenames() -> List[str]:
        """
        Returns a list of the filenames of the libraries of this type in the tdw module.
        """

        raise Exception()

    def get_record(self, name: str) -> Optional[T]:
        """
        Returns a record with the specified name. If that record can't be found, returns None.

        :param name: The name of the record.
        """

        records = [r for r in self.records if r.name == name]

        if len(records) == 0:
            return None
        else:
            return records[0]

    def search_records(self, search: str) -> List[T]:
        """
        Returns a list of records whose names include the search keyword.

        :param search: The string to search for in the model name.
        """

        return [r for r in self.records if search in r.name]

    def add_or_update_record(self, record: T, overwrite: bool, write: bool = True, quiet: bool = True) -> bool:
        """
        Add a new record or update an existing record.

        :param record: The record.
        :param overwrite: If true, overwrite the record if it already exists.
        :param write: If true, write the library data to disk (overwriting the existing file).
        :param quiet: If true, silently correct the model name if need be.
        """

        # Valid the name of the record.
        name_ok, name, problems = self.get_valid_record_name(record.name, overwrite)
        record.name = name
        if not name_ok and not quiet:
            print(f"Renaming this record to {name} because:")
            for p in problems:
                print(f"\t{p}")

        added = False
        if len([r for r in self.records if r.name == record.name]) > 0:
            # If this record exists and we want to overwrite, update the record.
            if overwrite:
                records_list = [r for r in self.records if r.name != record.name]
                records_list.append(record)
                added = True
        # Add the record.
        else:
            self.records.append(record)
            added = True

        # Write to disk.
        if added:
            if record.name in self.data["records"]:
                self.data["records"][record.name] = record.get_serializable()
            else:
                self.data["records"].update({record.name: record.get_serializable()})
        if write:
            self.write()

        return added

    def remove_record(self, record: Union[str, T], write: bool = True) -> bool:
        """
        Remove a record. Returns true if the record was removed.

        :param record: The record or the name of the record.
        :param write: If true, write the library data to disk  (overwriting the existing file).
        """

        if isinstance(record, str):
            record_name = record
        else:
            record_name = record.name

        records_list = [r for r in self.records if r.name != record_name]
        removed = len(records_list) < len(self.records)
        if removed:
            del self.data["records"][record_name]
            self.records = records_list
        if write:
            self.write()

        return removed

    def write(self, pretty=True) -> None:
        """
        Write the data to disk.

        :param pretty: Pretty print.
        """

        with open(self.library, "wt") as f:
            if pretty:
                json.dump(self.data, f, sort_keys=True, indent=4)
            else:
                json.dump(self.data, f)

    def get_valid_record_name(self, name: str, overwrite: bool) -> Tuple[bool, str, List[str]]:
        """
        Generates a valid record name. Returns: true if the name is good as-is, the new name, and a list of problems with the old name.

        :param name: The name of a record we'd like to add.
        :param overwrite: If true, raise an exception if the record doesn't exist. Otherwise, overwrite. If False: If the record exists, suggest a new name.
        """

        record_names = [r.name for r in self.records]

        if overwrite and name not in record_names:
            return False, name, [f"Can't override a record named {name} because no such record exists!"]

        good_name = name[:]
        ok = True
        problems: List[str] = []
        good_name = good_name.replace(" ", "_")
        if good_name != name:
            ok = False
            problems.append("Name has spaces. They have been replaced with underscores.")
        good_name = good_name.lower()
        if good_name != name:
            ok = False
            problems.append("Name has uppercase letters. They are now all lowercase.")

        if not overwrite and good_name in record_names:
            ok = False
            while good_name in record_names:
                good_name = good_name + token_hex(2)
            problems.append(f"A record named {name} already exists, and we don't want to overwrite it.")
        return ok, good_name, problems

    def _generate_record(self, data: dict) -> T:
        """
        Generate a record of type T from JSON data.

        :param data: The record JSON data.
        """

        raise Exception("Not defined.")


class ModelLibrarian(_Librarian[ModelRecord]):
    """
    Librarian class for model metadata.
    """

    def get_model_wnids_and_wcategories(self) -> Dict[str, str]:
        """
        Returns a dictionary of all model wnids and categories.
        Key=wnid Value=category
        """

        wnids: Dict[str, str] = {}
        for model in self.records:
            if model.wnid in wnids:
                if wnids[model.wnid] != model.wcategory:
                    print(f"WARNING: Model {model.name} wcategory is {model.wcategory} (expected: {wnids[model.wnid]})")
            else:
                wnids.update({model.wnid: model.wcategory})
        return wnids

    def get_model_wnids(self) -> List[str]:
        """
        Returns a list of all unique wnids in the database, sorted numerically.
        """

        return sorted(set([r.wnid for r in self.records]))

    def get_all_models_in_wnid(self, wnid: str) -> List[ModelRecord]:
        """
        Returns a list of all models with the same wnid.

        :param wnid: The WordNet ID.
        """

        return [r for r in self.records if r.wnid == wnid]

    def get_flex_models(self) -> List[ModelRecord]:
        """
        Returns a list of all Flex-compatible models.
        """

        return [r for r in self.records if r.flex]

    @staticmethod
    def get_library_filenames() -> List[str]:
        return ["models_core.json", "models_full.json", "models_special.json", "models_flex.json"]

    def _generate_record(self, data: dict) -> T:
        return ModelRecord(data)


class MaterialLibrarian(_Librarian[MaterialRecord]):
    """
    Librarian class for material metadata.
    """

    def get_all_materials_of_type(self, material_type: str) -> List[MaterialRecord]:
        """
        Returns a list of all material records of a given type.

        :param material_type: The type of material.
        """

        return [r for r in self.records if r.type == material_type]

    def get_material_types(self) -> List[str]:
        """
        Returns a list of all types of materials, sorted alphabetically.
        """

        return sorted(set([r.type for r in self.records]))

    @staticmethod
    def get_library_filenames() -> List[str]:
        return ["materials_med.json", "materials_low.json", "materials_high.json"]

    def _generate_record(self, data: dict) -> T:
        return MaterialRecord(data)


class SceneLibrarian(_Librarian[SceneRecord]):
    """
    Librarian class for scene metadata.
    """

    @staticmethod
    def get_library_filenames() -> List[str]:
        return ["scenes.json"]

    def _generate_record(self, data: dict) -> T:
        return SceneRecord(data)


class HDRISkyboxLibrarian(_Librarian[HDRISkyboxRecord]):
    """
    Librarian class for HDRI skybox metadata.
    """

    @staticmethod
    def get_library_filenames() -> List[str]:
        return ["hdri_skyboxes.json"]

    def _generate_record(self, data: dict) -> T:
        return HDRISkyboxRecord(data)


class HumanoidAnimationLibrarian(_Librarian[HumanoidAnimationRecord]):
    """
    Librarian class for humanoid animation metadata.
    """

    @staticmethod
    def get_library_filenames() -> List[str]:
        return ["humanoid_animations.json"]

    def _generate_record(self, data: dict) -> T:
        return HumanoidAnimationRecord(data)


class HumanoidLibrarian(_Librarian[HumanoidRecord]):
    """
    Librarian class for humanoid metadata.
    """

    @staticmethod
    def get_library_filenames() -> List[str]:
        return ["humanoids.json"]

    def _generate_record(self, data: dict) -> T:
        return HumanoidRecord(data)



import numpy as np
import math
import base64
import json
import scipy.signal as sg
from enum import Enum
from typing import Dict, Optional, Tuple, Union, List
from tdw.output_data import OutputData, Rigidbodies, Collision, EnvironmentCollision
from pathlib import Path
from pkg_resources import resource_filename
from csv import DictReader
import io


class AudioMaterial(Enum):
    """
    These are the materials currently supported for impact sounds in pyImpact.  More will be added in time.
    """

    ceramic = 0
    glass = 1
    metal = 2
    hardwood = 3
    wood = 4
    cardboard = 5


class ObjectInfo:
    """
    Impact sound data for an object in a TDW model library.
    The audio values here are just recommendations; you can apply different values if you want.
    """

    def __init__(self, name: str, amp: float, mass: float, material: AudioMaterial, library: str, bounciness: float):
        """
        :param name: The model name.
        :param amp: The sound amplitude.
        :param mass: The object mass.
        :param material: The audio material.
        :param library: The path to the model library (see ModelLibrarian documentation).
        :param bounciness: The bounciness value for a Unity physics material.
        """

        self.amp = amp
        self.library = library
        self.mass = mass
        self.material = material
        self.name = name
        self.bounciness = bounciness


# Density per audio material.
DENSITIES: Dict[AudioMaterial, float] = {AudioMaterial.ceramic: 2180,
                                         AudioMaterial.glass: 2500,
                                         AudioMaterial.metal: 8450,
                                         AudioMaterial.hardwood: 900,
                                         AudioMaterial.wood: 690,
                                         AudioMaterial.cardboard: 698}


class Base64Sound:
    """
    A sound encoded as a base64 string.

    Fields:

    - `bytes` Byte data of the sound.
    - `wav_str` base64 string of the sound. Send this to the build.
    - `length` The length of the byte array.
    """

    def __init__(self, snd: np.array):
        """
        :param snd: The sound byte array.
        """

        tst1 = np.array(snd * 32767, dtype='int16')
        tst2 = bytes(tst1)
        tst3 = base64.b64encode(tst2).decode('utf-8')

        self.bytes = tst2
        self.wav_str = tst3
        self.length = len(tst2)


class Modes:
    """
    Resonant mode properties: Frequencies, powers, and times.
    """

    def __init__(self, frequencies: np.array, powers: np.array, decay_times: np.array):
        """
        :param frequencies: numpy array of mode frequencies in Hz
        :param powers: numpy array of mode onset powers in dB re 1.
        :param decay_times: numpy array of mode decay times (i.e. the time in ms it takes for each mode to decay 60dB from its onset power)
        """
        self.frequencies = frequencies
        self.powers = powers
        self.decay_times = decay_times

    def sum_modes(self, fs: int = 44100) -> np.array:
        """
        Create mode time-series from mode properties and sum them together.

        :return A synthesized sound.
        """

        # Scroll through modes
        for i in range(len(self.frequencies)):
            H_dB = 80 + self.powers[i]
            L_ms = self.decay_times[i] * H_dB / 60
            mLen = math.ceil(L_ms / 1e3 * fs)
            # if mode length is greater than the current time-series we have had make our time series longer
            max_len = mLen
            if mLen > max_len:
                max_len = mLen
            tt = np.arange(0, max_len) / fs
            # synthesize a sinusoid
            mode = np.cos(2 * math.pi * self.frequencies[i] * tt)
            mode = mode * (10 ** (self.powers[i] / 20))
            dcy = tt * (60 / (self.decay_times[i] / 1e3))
            env = 10 ** (-dcy / 20)
            mode = mode * env
            if i == 0:
                synth_sound = mode
            else:
                synth_sound = Modes.mode_add(synth_sound, mode)
        return synth_sound

    @staticmethod
    def mode_add(a: np.array, b: np.array) -> np.array:
        """
        Add together numpy arrays of different lengths by zero-padding the shorter.

        :param a: The first array.
        :param b: The second array.

        :return The summed modes.
        """

        if len(a) < len(b):
            c = b.copy()
            c[:len(a)] += a
        else:
            c = a.copy()
            c[:len(b)] += b
        return c


class CollisionInfo:
    """
    Class containing information about collisions required by pyImpact to determine the volume of impact sounds.
    """

    def __init__(self, obj1_modes: Modes, obj2_modes: Modes, amp: float = 0.5, init_speed: float = 1):
        """
        :param amp: Amplitude of the first collision (must be between 0 and 1).
        :param init_speed: The speed of the initial collision (all collisions will be scaled relative to this).
        :param obj1_modes: The object's modes.
        :param obj2_modes: The other object's modes.
        """

        self.count = 0
        self.amp = amp
        # The speed of the initial collision.
        self.init_speed = init_speed
        # The audio modes.
        self.obj1_modes = obj1_modes
        self.obj2_modes = obj2_modes

    def count_collisions(self) -> None:
        """
        Update the counter for how many times two objects have collided.
        """

        self.count += 1


class CollisionType(Enum):
    """
    The "type" of a collision, defined by the motion of the object.

    none = No collision
    impact = The object "entered" a collision
    scrape = The object "stayed" in a collision with a low angular velocity.
    roll = The object "stayed" in a collision with a high angular velocity.
    """

    none = 1,
    impact = 2,
    scrape = 4,
    roll = 8


class CollisionTypesOnFrame:
    """
    All types of collision (impact, scrape, roll, none) between an object and any other objects or the environment on this frame.

    Usage:

    ```python
    from tdw.controller import Controller
    from tdw.py_impact import CollisionTypesOnFrame

    object_id = c.get_unique_id()
    c = Controller()
    c.start()

    # Your code here.

    # Request the required output data (do this at the start of the simulation, not per frame).
    resp = c.communicate([{"$type": "send_collisions",
                           "enter": True,
                           "exit": False,
                           "stay": True,
                           "collision_types": ["obj", "env"]},
                          {"$type": "send_rigidbodies",
                           "frequency": "always"}])

    # Parse the output data and get collision type data.
    ctof = CollisionTypesOnFrame(object_id, resp)

    # Read the dictionaries of collidee IDs and collision types.
    for collidee_id in ctof.collisions:
        collision_type = ctof.collisions[collidee_id]
        print(collidee_id, collision_type)

    # Check the environment collision.
    print(ctof.env_collision_type)
    ```
    """

    def __init__(self, object_id: int, resp: List[bytes]):
        """
        :param object_id: The unique ID of the colliding object.
        :param resp: The response from the build.
        """

        collisions, env_collisions, rigidbodies = PyImpact.get_collisions(resp)

        # The type of collision with each collidee.
        self.collisions: Dict[int, CollisionType] = dict()
        # The type of environment collision, if any.
        self.env_collision = CollisionType.none

        # If there is no Rigidbodies output data, then nothing can be parsed.
        if rigidbodies is None:
            return

        # Get the rigidbody data for this object.
        for i in range(rigidbodies.get_num()):
            if rigidbodies.get_id(i) == object_id:
                # Get the angular velocity of this object.
                ang_vel = rigidbodies.get_angular_velocity(i)

                # My collisions with other objects on this frame.
                # Key = the collidee ID. Value = list of states.
                my_collisions: Dict[int, List[str]] = dict()

                for co in collisions:
                    if co.get_collider_id() == object_id:
                        collidee = co.get_collidee_id()
                        if collidee not in my_collisions:
                            my_collisions.update({collidee: []})
                        my_collisions[collidee].append(co.get_state())
                # Get the collision type.
                for collidee in my_collisions:
                    self.collisions[collidee] = self._get_collision_type(ang_vel, my_collisions[collidee])
                env_collision_states: List[str] = []
                for co in env_collisions:
                    env_collision_states.append(co.get_state())

    @staticmethod
    def _get_collision_type(ang_vel: tuple, states: List[str]) -> CollisionType:
        """
        :param ang_vel: The angular velocity of this object.

        :param states: The states of all collisions experienced by this object on this frame.

        :return: The type of collision that the object is experiencing.
        """

        # If there is any "enter" state in this frame, then it is an impact.
        if "enter" in states:
            return CollisionType.impact
        # If there are "stay" state(s) but no "enter" state, then it is either a roll or a scrape.
        elif "stay" in states:
            # If there is a high angular velocity, then it is a roll.
            if np.linalg.norm(ang_vel) > 1:
                return CollisionType.roll
            # Otherwise, it's a scrape.
            else:
                return CollisionType.scrape
        else:
            return CollisionType.none


class PyImpact:
    """
    Generate impact sounds from physics data.

    Sounds are synthesized as described in: [Traer,Cusimano and McDermott, A PERCEPTUALLY INSPIRED GENERATIVE MODEL OF RIGID-BODY CONTACT SOUNDS, Digital Audio Effects, (DAFx), 2019](http://dafx2019.bcu.ac.uk/papers/DAFx2019_paper_57.pdf)

    For a general guide on impact sounds in TDW, read [this](../misc_frontend/impact_sounds.md).

    Usage:

    ```python
    from tdw.controller import Controller
    from tdw.py_impact import PyImpact

    p = PyImpact()
    c = Controller()
    c.start()

    # Your code here.

    c.communicate(p.get_impact_sound_command(arg1, arg2, ... ))
    ```
    """

    def __init__(self, initial_amp: float = 0.5, prevent_distortion: bool = True):
        """
        :param initial_amp: The initial amplitude, i.e. the "master volume". Must be > 0 and < 1.
        :param prevent_distortion: If True, clamp amp values to <= 0.99
        """

        assert 0 < initial_amp < 1, f"initial_amp is {initial_amp} (must be > 0 and < 1)."

        self.initial_amp = initial_amp
        self.prevent_distortion = prevent_distortion

        # The collision info per set of objects.
        self.object_modes: Dict[int, Dict[int, CollisionInfo]] = {}

        # Cache the material data. This is use to reset the material modes.
        self.material_data: Dict[str, dict] = {}
        for mat, path in zip(["ceramic", "hardwood", "metal", "glass", "wood", "cardboard"],
                             ["Ceramic_mm", "Poplar_mm", "MetalStrip_mm", "Mirror_mm", "BalsaWood_mm", "Cardboard_mm"]):
            # Load the JSON data.
            data = json.loads(Path(resource_filename(__name__, f"py_impact/material_data/{path}.json")).read_text())
            self.material_data.update({mat: data})

    def _get_object_modes(self, material: Union[str, AudioMaterial]) -> Modes:
        """
        :param material: The audio material.

        :return: The audio modes.
        """

        data = self.material_data[material] if isinstance(material, str) else self.material_data[material.name]
        # Load the mode properties.
        f = -1
        p = -1
        t = -1
        for jm in range(0, 10):
            jf = 0
            while jf < 20:
                jf = data["cf"][jm] + np.random.normal(0, data["cf"][jm] / 10)
            jp = data["op"][jm] + np.random.normal(0, 10)
            jt = 0
            while jt < 0.001:
                jt = data["rt"][jm] + np.random.normal(0, data["rt"][jm] / 10)
            if jm == 0:
                f = jf
                p = jp
                t = jt * 1e3
            else:
                f = np.append(f, jf)
                p = np.append(p, jp)
                t = np.append(t, jt * 1e3)
        return Modes(f, p, t)

    def get_sound(self, collision: Union[Collision, EnvironmentCollision], rigidbodies: Rigidbodies, id1: int, mat1: str, id2: int, mat2: str, amp2re1: float) -> Optional[Base64Sound]:
        """
        Produce sound of two colliding objects as a byte array.

        :param collision: TDW `Collision` or `EnvironmentCollision` output data.
        :param rigidbodies: TDW `Rigidbodies` output data.
        :param id1: The object ID for one of the colliding objects.
        :param mat1: The material label for one of the colliding objects.
        :param id2: The object ID for the other object.
        :param mat2: The material label for the other object.
        :param amp2re1: The sound amplitude of object 2 relative to that of object 1.

        :return Sound data as a Base64Sound object.
        """

        # Set the object modes.
        if id2 not in self.object_modes:
            self.object_modes.update({id2: {}})
        if id1 not in self.object_modes[id2]:
            self.object_modes[id2].update({id1: CollisionInfo(self._get_object_modes(mat2),
                                                              self._get_object_modes(mat1),
                                                              amp=self.initial_amp)})

        obj_col = isinstance(collision, Collision)

        # Unpack useful parameters.
        # Compute normal velocity at impact.
        vel = 0
        if obj_col:
            vel = collision.get_relative_velocity()
        else:
            for i in range(rigidbodies.get_num()):
                if rigidbodies.get_id(i) == id2:
                    vel = rigidbodies.get_velocity(i)
                    break
        vel = np.asarray(vel)
        speed = np.square(vel)
        speed = np.sum(speed)
        speed = math.sqrt(speed)
        nvel = vel / np.linalg.norm(vel)
        num_contacts = collision.get_num_contacts()
        nspd = []
        for jc in range(0, num_contacts):
            tmp = np.asarray(collision.get_contact_normal(jc))
            tmp = tmp / np.linalg.norm(tmp)
            tmp = np.arccos(np.clip(np.dot(tmp, nvel), -1.0, 1.0))
            # Scale the speed by the angle (i.e. we want speed Normal to the surface).
            tmp = speed * np.cos(tmp)
            nspd.append(tmp)
        normal_speed = np.mean(nspd)
        # Get indices of objects in collisions
        id1_index = None
        id2_index = None
        for i in range(rigidbodies.get_num()):
            if rigidbodies.get_id(i) == id1:
                id1_index = i
            if rigidbodies.get_id(i) == id2:
                id2_index = i

        # Make sure both IDs were found. If they aren't, don't return a sound.
        if obj_col and (id1_index is None or id2_index is None):
            return None

        m1 = rigidbodies.get_mass(id1_index) if obj_col else 1000
        m2 = rigidbodies.get_mass(id2_index)
        mass = np.min([m1, m2])

        # Re-scale the amplitude.
        if self.object_modes[id2][id1].count == 0:
            # Sample the modes.
            sound, modes_1, modes_2 = self.make_impact_audio(amp2re1, mass, mat1=mat1, mat2=mat2, id1=id1, id2=id2)
            # Save collision info - we will need for later collisions.
            amp = self.object_modes[id2][id1].amp
            self.object_modes[id2][id1].init_speed = normal_speed
            self.object_modes[id2][id1].obj1_modes = modes_1
            self.object_modes[id2][id1].obj2_modes = modes_2
        else:
            amp = self.object_modes[id2][id1].amp * normal_speed / self.object_modes[id2][id1].init_speed
            # Adjust modes here so that two successive impacts are not identical.
            modes_1 = self.object_modes[id2][id1].obj1_modes
            modes_2 = self.object_modes[id2][id1].obj2_modes
            modes_1.powers = modes_1.powers + np.random.normal(0, 2, len(modes_1.powers))
            modes_2.power = modes_2.powers + np.random.normal(0, 2, len(modes_1.powers))
            sound = PyImpact.synth_impact_modes(modes_1, modes_2, mass)
            self.object_modes[id2][id1].obj1_modes = modes_1
            self.object_modes[id2][id1].obj2_modes = modes_2

        # On rare occasions, it is possible for PyImpact to fail to generate a sound.
        if sound is None:
            return None

        # Count the collisions.
        self.object_modes[id2][id1].count_collisions()

        # Prevent distortion by clamping the amp.
        if self.prevent_distortion and amp > 0.99:
            amp = 0.99

        sound = amp * sound / np.max(np.abs(sound))
        return Base64Sound(sound)

    def get_impact_sound_command(self, collision: Union[Collision, EnvironmentCollision], rigidbodies: Rigidbodies, target_id: int, target_mat: str, target_amp: float, other_id: int, other_mat: str, other_amp: float, play_audio_data: bool = True) -> dict:
        """
        Create an impact sound, and return a valid command to play audio data in TDW.
        "target" should usually be the smaller object, which will play the sound.
        "other" should be the larger (stationary) object.

        :param collision: TDW `Collision` or `EnvironmentCollision` output data.
        :param target_amp: The target's amp value.
        :param target_mat: The target's audio material.
        :param other_amp: The other object's amp value.
        :param other_id: The other object's ID.
        :param other_mat: The other object's audio material.
        :param rigidbodies: TDW `Rigidbodies` output data.
        :param target_id: The ID of the object that will play the sound.
        :param play_audio_data: If True, return a `play_audio_data` command. If False, return a `play_point_source_data` command (useful only with Resonance Audio; see Command API).

        :return A `play_audio_data` or `play_point_source_data` command that can be sent to the build via `Controller.communicate()`.
        """

        amp2re1 = other_amp / target_amp

        impact_audio = self.get_sound(collision, rigidbodies, other_id, other_mat, target_id, target_mat, amp2re1)
        if impact_audio is not None:
            return {"$type": "play_audio_data" if play_audio_data else "play_point_source_data",
                    "id": target_id,
                    "num_frames": impact_audio.length,
                    "num_channels": 1,
                    "frame_rate": 44100,
                    "wav_data": impact_audio.wav_str,
                    "y_pos_offset": 0.1}
        # If PyImpact failed to generate a sound (which is rare!), fail silently here.
        else:
            return {"$type": "do_nothing"}

    def make_impact_audio(self, amp2re1: float, mass: float, id1: int, id2: int, mat1: str = 'cardboard', mat2: str = 'cardboard') -> (np.array, Modes, Modes):
        """
        Generate an impact sound.

        :param mat1: The material label for one of the colliding objects.
        :param mat2: The material label for the other object.
        :param amp2re1: The sound amplitude of object 2 relative to that of object 1.
        :param mass: The mass of the smaller of the two colliding objects.
        :param id1: The ID for the one of the colliding objects.
        :param id2: The ID for the other object.

        :return The sound, and the object modes.
        """

        # Unpack material names.
        for jmat in range(0, len(AudioMaterial)):
            if mat1 == AudioMaterial(jmat):
                tmp1 = AudioMaterial(jmat)
                mat1 = tmp1.name
            if mat2 == AudioMaterial(jmat):
                tmp2 = AudioMaterial(jmat)
                mat2 = tmp2.name
        # Sample modes of object1.
        modes_1 = self.object_modes[id2][id1].obj1_modes
        modes_2 = self.object_modes[id2][id1].obj2_modes
        # Scale the two sounds as specified.
        modes_2.decay_times = modes_2.decay_times + 20 * np.log10(amp2re1)
        snth = PyImpact.synth_impact_modes(modes_1, modes_2, mass)
        return snth, modes_1, modes_2

    @staticmethod
    def get_impulse_response(modes1: Modes, modes2: Modes) -> np.array:
        """
        Generate an impulse response from specified modes for two objects.

        :param modes1: Modes of object 1. A numpy array with: column1=mode frequencies (Hz); column2=mode onset powers in dB; column3=mode RT60s in milliseconds;
        :param modes2: Modes of object 2. Formatted as modes1/modes2.

        :return The impulse response.
        """
        h1 = modes1.sum_modes()
        h2 = modes2.sum_modes()
        h = Modes.mode_add(h1, h2)
        return h

    @staticmethod
    def synth_impact_modes(modes1: Modes, modes2: Modes, mass: float) -> np.array:
        """
        Generate an impact sound from specified modes for two objects, and the mass of the smaller object.

        :param modes1: Modes of object 1. A numpy array with: column1=mode frequencies (Hz); column2=mode onset powers in dB; column3=mode RT60s in milliseconds;
        :param modes2: Modes of object 2. Formatted as modes1/modes2.
        :param mass: the mass of the smaller of the two colliding objects.

        :return The impact sound.
        """

        h = PyImpact.get_impulse_response(modes1, modes2)
        if len(h) == 0:
            return None
        # Convolve with force, with contact time scaled by the object mass.
        max_t = 0.001 * mass
        # A contact time over 2ms is unphysically long.
        max_t = np.min([max_t, 2e-3])
        n_pts = int(np.ceil(max_t * 44100))
        tt = np.linspace(0, np.pi, n_pts)
        frc = np.sin(tt)
        x = sg.fftconvolve(h, frc)
        x = x / abs(np.max(x))
        return x

    @staticmethod
    def get_object_info(csv_file: Union[str, Path] = "") -> Dict[str, ObjectInfo]:
        """
        Returns ObjectInfo values.
        As of right now, only a few objects in the TDW model libraries are included. More will be added in time.

        :param csv_file: The path to the .csv file containing the object info. By default, it will load `tdw/py_impact/objects.csv`. If you want to make your own spreadsheet, use this file as a reference.

        :return: A list of default ObjectInfo. Key = the name of the model. Value = object info.
        """

        objects: Dict[str, ObjectInfo] = {}
        # Load the objects.csv metadata file.
        if isinstance(csv_file, str):
            # Load the default file.
            if csv_file == "":
                csv_file = str(Path(resource_filename(__name__, f"py_impact/objects.csv")).resolve())
            else:
                csv_file = str(Path(csv_file).resolve())
        else:
            csv_file = str(csv_file.resolve())

        # Parse the .csv file.
        with io.open(csv_file, newline='', encoding='utf-8-sig') as f:
            reader = DictReader(f)
            for row in reader:
                o = ObjectInfo(name=row["name"], amp=float(row["amp"]), mass=float(row["mass"]),
                               material=AudioMaterial[row["material"]], library=row["library"],
                               bounciness=float(row["bounciness"]))
                objects.update({o.name: o})

        return objects

    @staticmethod
    def get_collisions(resp: List[bytes]) -> Tuple[List[Collision], List[EnvironmentCollision], Optional[Rigidbodies]]:
        """
        Parse collision and rigibody data from the output data.

        :param resp: The response from the build.

        :return: A list of collisions on this frame (can be empty), a list of environment collisions on this frame (can be empty), and Rigidbodies data (can be `None`).
        """

        if len(resp) == 1:
            return [], [], None
        collisions: List[Collision] = []
        environment_collisions: List[EnvironmentCollision] = []
        rigidbodies: Optional[Rigidbodies] = None
        for r in resp[:-1]:
            r_id = OutputData.get_data_type_id(r)
            if r_id == 'coll':
                collisions.append(Collision(r))
            if r_id == 'rigi':
                rigidbodies = Rigidbodies(r)
            if r_id == 'enco':
                environment_collisions.append(EnvironmentCollision(r))

        return collisions, environment_collisions, rigidbodies

    @staticmethod
    def is_valid_collision(collision: Union[Optional[Collision], Optional[EnvironmentCollision]]) -> bool:
        """
        :param collision: Collision or EnvironmentCollision output data from the build.

        :return: True if this collision can be used to generate an impact sound.
        """

        return collision is not None and ((isinstance(collision, Collision) and
                                           np.linalg.norm(collision.get_relative_velocity()) > 0) or
                                          isinstance(collision, EnvironmentCollision))

    def reset(self, initial_amp: float = 0.5) -> None:
        """
        Reset PyImpact. This is somewhat faster than creating a new PyImpact object per trial.

        :param initial_amp: The initial amplitude, i.e. the "master volume". Must be > 0 and < 1.
        """

        assert 0 < initial_amp < 1, f"initial_amp is {initial_amp} (must be > 0 and < 1)."

        # Clear the object data.
        self.object_modes.clear()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class Raycast(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsRaycast(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = Raycast()
        x.Init(buf, n + offset)
        return x

    # Raycast
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # Raycast
    def Hit(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return bool(self._tab.Get(tdw.flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Raycast
    def RaycastId(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Raycast
    def ObjectId(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Raycast
    def Normal(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Raycast
    def Point(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def RaycastStart(builder): builder.StartObject(5)
def RaycastAddHit(builder, hit): builder.PrependBoolSlot(0, hit, 0)
def RaycastAddRaycastId(builder, raycastId): builder.PrependInt32Slot(1, raycastId, 0)
def RaycastAddObjectId(builder, objectId): builder.PrependInt32Slot(2, objectId, 0)
def RaycastAddNormal(builder, normal): builder.PrependStructSlot(3, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(normal), 0)
def RaycastAddPoint(builder, point): builder.PrependStructSlot(4, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(point), 0)
def RaycastEnd(builder): return builder.EndObject()



from tdw.controller import Controller
from tdw.librarian import ModelRecord
from typing import Tuple, List
from tdw.output_data import LogMessage
from tdw.backend import paths
from subprocess import Popen
import json
from pathlib import Path
from argparse import ArgumentParser
from tdw.model_pipeline.missing_materials import MissingMaterials


class Validator(Controller):
    """
    Check a model asset bundle for problems.
    First, load the model into the build and report back any errors.
    Then, scan the model for missing materials.

    This controller will attempt to launch a build.
    If one doesn't exist at the expected location, you'll need to launch one manually.
    """

    PINK_MATERIALS_REPORT: List[str] = ["This model has missing (pink) materials."]

    def __init__(self, record_path: str, asset_bundle_path: str, port=1071):
        """
        :param record_path: The path to the temporary record file.
        :param asset_bundle_path: The path to the local asset bundle.
        :param port: The port.
        """

        self.record_path: str = record_path
        self.record: ModelRecord = ModelRecord(json.loads(Path(record_path).read_text(encoding="utf-8")))
        self.asset_bundle_path: str = asset_bundle_path
        if not self.asset_bundle_path.startswith("file:///"):
            self.asset_bundle_path = "file:///" + self.asset_bundle_path

        super().__init__(port)

    def run(self, quiet=True) -> Tuple[bool, List[str]]:
        """
        Run verification. First, test for general errors with the model. Next, test for missing materials.

        :param quiet: If true, suppress messages.
        """

        MissingMaterials.start(self)

        # Run the verifier.
        if not quiet:
            print("Running verifier...")

        resp = self.communicate({"$type": "send_model_report",
                                 "name": self.record.name,
                                 "url": self.asset_bundle_path,
                                 "id": self.get_unique_id(),
                                 "scale_factor": self.record.scale_factor})
        reports = []
        for r in resp[:-1]:
            msg = LogMessage(r)
            reports.append(msg.get_message_type() + ": " + msg.get_message())

        # Check for missing materials.
        if not quiet:
            print("Checking for missing materials...")

        pink_materials = MissingMaterials.materials_are_missing(self, self.record, self.asset_bundle_path)
        if pink_materials:
            reports.append(self.PINK_MATERIALS_REPORT)

        self.kill_build()
        return len(reports) == 0, reports

    def kill_build(self):
        """
        Kill the build process.
        """

        self.communicate({"$type": "terminate"})


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--record_path", type=str, help="The path to the temporary record file.")
    parser.add_argument("--asset_bundle_path", type=str, help="The path to the local asset bundle.")
    args = parser.parse_args()

    # Get the physics quality.
    c = Validator(record_path=args.record_path,
                  asset_bundle_path=args.asset_bundle_path)
    ok, reports = c.run()
    report = {"ok": ok,
              "reports": reports}

    root_dir = paths.ASSET_BUNDLE_VERIFIER_OUTPUT_DIR
    if not root_dir.exists():
        root_dir.mkdir(parents=True)
    paths.VALIDATOR_REPORT_PATH.write_text(json.dumps(report), encoding="utf-8")



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class AudioSource(object):
    __slots__ = ['_tab']

    # AudioSource
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # AudioSource
    def Id(self): return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, self._tab.Pos + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # AudioSource
    def IsPlaying(self): return self._tab.Get(tdw.flatbuffers.number_types.BoolFlags, self._tab.Pos + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(4))

def CreateAudioSource(builder, id, isPlaying):
    builder.Prep(4, 8)
    builder.Pad(3)
    builder.PrependBool(isPlaying)
    builder.PrependInt32(id)
    return builder.Offset()



from tdw.controller import Controller
from tdw.librarian import ModelLibrarian, ModelRecord
from tdw.output_data import Images
from tdw.backend.paths import ASSET_BUNDLE_VERIFIER_OUTPUT_DIR
from PIL import Image
from io import BytesIO
from tdw.tdw_utils import TDWUtils
from math import radians, sin, cos
from tqdm import tqdm
import io


class MissingMaterials:
    """
    Check for missing materials.
    """

    HELL_PINK = (255, 0, 255)
    DELTA_THETA = 15
    OUTPUT_FILE = ASSET_BUNDLE_VERIFIER_OUTPUT_DIR.joinpath("missing_materials.txt")
    if not ASSET_BUNDLE_VERIFIER_OUTPUT_DIR.exists():
        ASSET_BUNDLE_VERIFIER_OUTPUT_DIR.mkdir(parents=True)
    OBJECT_ID = 0

    @staticmethod
    def start(c: Controller) -> None:
        """
        Start the controller.

        :param c: The controller.
        """

        c.start()

        c.communicate([{"$type": "simulate_physics",
                        "value": False},
                       {"$type": "create_empty_environment"},
                       {"$type": "set_render_quality",
                        "render_quality": 0},
                       {"$type": "set_post_process",
                        "value": False},
                       {"$type": "create_avatar",
                        "id": "a",
                        "type": "A_Img_Caps_Kinematic"},
                       {"$type": "set_pass_masks",
                        "pass_masks": ["_img"]}])

    @staticmethod
    def materials_are_missing(c: Controller, record: ModelRecord, url: str) -> bool:
        """
        Check for missing materials.

        :param c: The controller.
        :param record: The model metadata record.
        :param url: The path to the model.

        :return: True if the model is missing materials.
        """

        # Disable post-processing.
        # Disable shadows.
        # Create the avatar.
        # Set pass masks.
        # Send images.
        # Add the object.
        # Add the object.
        # Load the bounds of the model.
        scale = TDWUtils.get_unit_scale(record)
        c.communicate([{"$type": "send_images",
                        "frequency": "always"},
                       {"$type": "add_object",
                        "name": record.name,
                        "url": url,
                        "scale_factor": record.scale_factor,
                        "id": MissingMaterials.OBJECT_ID},
                       {"$type": "scale_object",
                        "id": MissingMaterials.OBJECT_ID,
                        "scale_factor": {"x": scale, "y": scale, "z": scale}}])
        x = 1.75
        y = 0.5
        z = 0

        # Equatorial orbit.
        theta = 0
        while theta < 360:
            # Get the new position.
            rad = radians(theta)
            x1 = cos(rad) * x - sin(rad) * z
            z1 = sin(rad) * x + cos(rad) * z

            if MissingMaterials._orbit_and_capture_image(c, x1, y, z1):
                return True
            theta += MissingMaterials.DELTA_THETA

        # Polar orbit.
        theta = 0
        while theta < 360:
            # Get the new position.
            rad = radians(theta)
            x1 = cos(rad) * x - sin(rad) * z
            y1 = (sin(rad) * x + cos(rad) * z) + y

            if MissingMaterials._orbit_and_capture_image(c, x1, y1, z):
                return True
            theta += MissingMaterials.DELTA_THETA

        return False

    @staticmethod
    def _orbit_and_capture_image(c: Controller, x: float, y: float, z: float) -> bool:
        """
        Orbit around an object and check for missing materials.

        :param c: The controller.
        :param x: The next x positional coordinate.
        :param y: The next y positional coordinate.
        :param z: The next z positional coordinate.

        :return: True if there is a missing material in the image, False otherwise.
        """

        # Teleport the avatar.
        # Look at the model.
        # Receive an image.
        resp = c.communicate([{"$type": "teleport_avatar_to",
                               "position": {"x": x, "y": y, "z": z}},
                              {"$type": "look_at",
                               "object_id": MissingMaterials.OBJECT_ID,
                               "use_centroid": True}])

        # Check if there are any pink pixels.
        images = Images(resp[0])

        img = Image.open(BytesIO(images.get_image(0)))

        for c in img.getcolors(maxcolors=100000):
            if c[1] == MissingMaterials.HELL_PINK:
                return True
        return False

    @staticmethod
    def run(c: Controller) -> None:
        """
        Check every model for missing materials.

        :param c: The controller.
        """

        # Create a new output file.
        if MissingMaterials.OUTPUT_FILE.exists():
            MissingMaterials.OUTPUT_FILE.unlink()
        MissingMaterials.OUTPUT_FILE = str(MissingMaterials.OUTPUT_FILE.resolve())

        MissingMaterials.start(c)

        print(f"The names of models with missing materials will be saved to: {MissingMaterials.OUTPUT_FILE}")
        for library_path in ModelLibrarian.get_library_filenames():
            print(library_path)
            lib = ModelLibrarian(library=library_path)
            pbar = tqdm(total=len(lib.records))
            for record in lib.records:
                if record.do_not_use:
                    pbar.update(1)
                    continue
                pbar.set_description(record.name)
                # Check for missing materials.
                if MissingMaterials.materials_are_missing(c, record, record.get_url()):
                    with io.open(MissingMaterials.OUTPUT_FILE, "at") as f:
                        f.write("\n" + record.name)
                # Cleanup.
                c.communicate([{"$type": "destroy_object",
                                "id": MissingMaterials.OBJECT_ID},
                               {"$type": "unload_asset_bundles"}])
                pbar.update(1)
            pbar.close()


if __name__ == "__main__":
    MissingMaterials.run(Controller())



from math import cos, sin, radians
from tdw.librarian import ModelRecord
from tdw.tdw_utils import TDWUtils
from tdw.model_pipeline.validator import Validator
from tdw.output_data import Images
from PIL import Image
from io import BytesIO
import numpy as np
from argparse import ArgumentParser
import json
from pathlib import Path


class PhysicsQualityWriter(Validator):
    """
    Assess the percentage of coverage by colliders on a model.
    Write the "physics quality" to the record.
    """

    def get_physics_quality(self) -> float:
        """
        Get the physics quality of the model.
        First, capture multiple images at different angles.
        Then, obscure the object with its own colliders and capture the same images.
        Compare the _id capture pass pixel count between each set of images.
        If the hull colliders obscure the entirety of the object in every image, physics_quality = 1.0
        """

        self.start()
        self.communicate([{"$type": "simulate_physics",
                           "value": False},
                          {"$type": "create_empty_environment"}])

        avatar_id = "A"
        delta_theta = 15

        print("Writing the physics quality to the record.")

        scale = TDWUtils.get_unit_scale(self.record)

        # Add the object.
        object_id = self.get_unique_id()
        self.communicate({"$type": "add_object",
                          "name": self.record.name,
                          "url": self.asset_bundle_path,
                          "scale_factor": self.record.scale_factor,
                          "id": object_id})

        # Scale the object to unit size.
        self.communicate({"$type": "scale_object",
                          "id": object_id,
                          "scale_factor": {"x": scale, "y": scale, "z": scale}})

        # Create the avatar.
        # Set the pass masks.
        self.communicate(TDWUtils.create_avatar(avatar_id=avatar_id))

        self.communicate({"$type": "set_pass_masks",
                          "avatar_id": avatar_id,
                          "pass_masks": ["_id"]})

        id_pass = []
        pink_pass = []
        for show_collider_hulls in [False, True]:
            x = 1.75
            y = 0.5
            z = 0

            # Show collider hulls.
            if show_collider_hulls:
                self.communicate({"$type": "show_collider_hulls",
                                  "id": object_id})

            # Reset the avatar.
            resp = self.communicate([{"$type": "teleport_avatar_to",
                                      "avatar_id": avatar_id,
                                      "position": {"x": x, "y": y, "z": z}},
                                     {"$type": "look_at",
                                      "avatar_id": avatar_id,
                                      "object_id": object_id,
                                      "use_centroid": True},
                                     {"$type": "send_images",
                                      "frequency": "always"}])

            # Equatorial orbit.
            theta = 0
            while theta < 360:
                # Get the number of pixels that aren't black.
                img = np.array(Image.open(BytesIO(Images(resp[0]).get_image(0))))
                grade = (256 * 256) - np.sum(np.all(img == np.array([0, 0, 0]), axis=2))

                if show_collider_hulls:
                    pink_pass.append(grade)
                else:
                    id_pass.append(grade)

                theta += delta_theta

                # Get the new position.
                rad = radians(theta)
                x1 = cos(rad) * x - sin(rad) * z
                z1 = sin(rad) * x + cos(rad) * z

                # Teleport the avatar.
                # Look at the object.
                resp = self.communicate([{"$type": "teleport_avatar_to",
                                          "avatar_id": avatar_id,
                                          "position": {"x": x1, "y": y, "z": z1}},
                                         {"$type": "look_at",
                                          "avatar_id": avatar_id,
                                          "object_id": object_id,
                                          "use_centroid": True}
                                         ])

            # Reset the avatar.
            resp = self.communicate([{"$type": "teleport_avatar_to",
                                      "avatar_id": avatar_id,
                                      "position": {"x": x, "y": y, "z": z}},
                                     {"$type": "look_at",
                                      "avatar_id": avatar_id,
                                      "object_id": object_id,
                                      "use_centroid": True},
                                     {"$type": "send_images",
                                      "frequency": "always"}])
            # Polar orbit.
            theta = 0
            while theta < 360:
                # Get the number of pixels that aren't black.
                img = np.array(Image.open(BytesIO(Images(resp[0]).get_image(0))))
                grade = (256 * 256) - np.sum(np.all(img == np.array([0, 0, 0]), axis=2))

                if show_collider_hulls:
                    pink_pass.append(grade)
                else:
                    id_pass.append(grade)

                theta += delta_theta

                # Get the new position.
                rad = radians(theta)
                x1 = cos(rad) * x - sin(rad) * z
                y1 = (sin(rad) * x + cos(rad) * z) + y

                # Teleport the avatar.
                # Look at the object.
                resp = self.communicate([{"$type": "teleport_avatar_to",
                                          "avatar_id": avatar_id,
                                          "position": {"x": x1, "y": y1, "z": 0}},
                                         {"$type": "look_at",
                                          "avatar_id": avatar_id,
                                          "object_id": object_id,
                                          "use_centroid": True}
                                         ])

        grades = [1 - (float(h) / float(i)) for h, i in zip(pink_pass, id_pass)]

        physics_quality = float(sum(grades)) / len(grades)
        print("Physics quality: " + str(physics_quality))

        # Kill the build.
        self.kill_build()
        return physics_quality


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--record_path", type=str, help="The path to the temporary record file")
    parser.add_argument("--asset_bundle_path", type=str, help="The path to the local asset bundle.")
    args = parser.parse_args()

    # Get the physics quality.
    c = PhysicsQualityWriter(record_path=args.record_path,
                             asset_bundle_path=args.asset_bundle_path)
    physics_quality = c.get_physics_quality()

    # Update the record.
    record_path = Path(args.record_path)
    record = ModelRecord(json.loads(record_path.read_text(encoding="utf-8")))
    record.physics_quality = physics_quality
    record_path.write_text(json.dumps(record.__dict__), encoding="utf-8")



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class AvatarStickyMitten(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsAvatarStickyMitten(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = AvatarStickyMitten()
        x.Init(buf, n + offset)
        return x

    # AvatarStickyMitten
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # AvatarStickyMitten
    def Id(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # AvatarStickyMitten
    def Position(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # AvatarStickyMitten
    def Rotation(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .Quaternion import Quaternion
            obj = Quaternion()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # AvatarStickyMitten
    def Forward(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # AvatarStickyMitten
    def Velocity(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # AvatarStickyMitten
    def AngularVelocity(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # AvatarStickyMitten
    def Mass(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Float32Flags, o + self._tab.Pos)
        return 0.0

    # AvatarStickyMitten
    def Sleeping(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return bool(self._tab.Get(tdw.flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # AvatarStickyMitten
    def RigidbodyParts(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            x = self._tab.Vector(o)
            x += tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j) * 36
            from .RigidbodyData import RigidbodyData
            obj = RigidbodyData()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # AvatarStickyMitten
    def RigidbodyPartsLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # AvatarStickyMitten
    def BodyParts(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            x = self._tab.Vector(o)
            x += tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j) * 44
            from .TransformData import TransformData
            obj = TransformData()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # AvatarStickyMitten
    def BodyPartsLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # AvatarStickyMitten
    def MittenCenterLeft(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            x = o + self._tab.Pos
            from .SimpleTransform import SimpleTransform
            obj = SimpleTransform()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # AvatarStickyMitten
    def MittenCenterRight(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            x = o + self._tab.Pos
            from .SimpleTransform import SimpleTransform
            obj = SimpleTransform()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # AvatarStickyMitten
    def HeldLeft(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, a + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # AvatarStickyMitten
    def HeldLeftAsNumpy(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.GetVectorAsNumpy(tdw.flatbuffers.number_types.Int32Flags, o)
        return 0

    # AvatarStickyMitten
    def HeldLeftLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # AvatarStickyMitten
    def HeldRight(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, a + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # AvatarStickyMitten
    def HeldRightAsNumpy(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.GetVectorAsNumpy(tdw.flatbuffers.number_types.Int32Flags, o)
        return 0

    # AvatarStickyMitten
    def HeldRightLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # AvatarStickyMitten
    def AnglesLeft(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(tdw.flatbuffers.number_types.Float32Flags, a + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # AvatarStickyMitten
    def AnglesLeftAsNumpy(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.GetVectorAsNumpy(tdw.flatbuffers.number_types.Float32Flags, o)
        return 0

    # AvatarStickyMitten
    def AnglesLeftLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # AvatarStickyMitten
    def AnglesRight(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(tdw.flatbuffers.number_types.Float32Flags, a + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # AvatarStickyMitten
    def AnglesRightAsNumpy(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.GetVectorAsNumpy(tdw.flatbuffers.number_types.Float32Flags, o)
        return 0

    # AvatarStickyMitten
    def AnglesRightLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def AvatarStickyMittenStart(builder): builder.StartObject(16)
def AvatarStickyMittenAddId(builder, id): builder.PrependUOffsetTRelativeSlot(0, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(id), 0)
def AvatarStickyMittenAddPosition(builder, position): builder.PrependStructSlot(1, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(position), 0)
def AvatarStickyMittenAddRotation(builder, rotation): builder.PrependStructSlot(2, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(rotation), 0)
def AvatarStickyMittenAddForward(builder, forward): builder.PrependStructSlot(3, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(forward), 0)
def AvatarStickyMittenAddVelocity(builder, velocity): builder.PrependStructSlot(4, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(velocity), 0)
def AvatarStickyMittenAddAngularVelocity(builder, angularVelocity): builder.PrependStructSlot(5, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(angularVelocity), 0)
def AvatarStickyMittenAddMass(builder, mass): builder.PrependFloat32Slot(6, mass, 0.0)
def AvatarStickyMittenAddSleeping(builder, sleeping): builder.PrependBoolSlot(7, sleeping, 0)
def AvatarStickyMittenAddRigidbodyParts(builder, rigidbodyParts): builder.PrependUOffsetTRelativeSlot(8, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(rigidbodyParts), 0)
def AvatarStickyMittenStartRigidbodyPartsVector(builder, numElems): return builder.StartVector(36, numElems, 4)
def AvatarStickyMittenAddBodyParts(builder, bodyParts): builder.PrependUOffsetTRelativeSlot(9, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(bodyParts), 0)
def AvatarStickyMittenStartBodyPartsVector(builder, numElems): return builder.StartVector(44, numElems, 4)
def AvatarStickyMittenAddMittenCenterLeft(builder, mittenCenterLeft): builder.PrependStructSlot(10, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(mittenCenterLeft), 0)
def AvatarStickyMittenAddMittenCenterRight(builder, mittenCenterRight): builder.PrependStructSlot(11, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(mittenCenterRight), 0)
def AvatarStickyMittenAddHeldLeft(builder, heldLeft): builder.PrependUOffsetTRelativeSlot(12, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(heldLeft), 0)
def AvatarStickyMittenStartHeldLeftVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def AvatarStickyMittenAddHeldRight(builder, heldRight): builder.PrependUOffsetTRelativeSlot(13, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(heldRight), 0)
def AvatarStickyMittenStartHeldRightVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def AvatarStickyMittenAddAnglesLeft(builder, anglesLeft): builder.PrependUOffsetTRelativeSlot(14, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(anglesLeft), 0)
def AvatarStickyMittenStartAnglesLeftVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def AvatarStickyMittenAddAnglesRight(builder, anglesRight): builder.PrependUOffsetTRelativeSlot(15, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(anglesRight), 0)
def AvatarStickyMittenStartAnglesRightVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def AvatarStickyMittenEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class VRRig(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsVRRig(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = VRRig()
        x.Init(buf, n + offset)
        return x

    # VRRig
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # VRRig
    def Position(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # VRRig
    def Rotation(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from .Quaternion import Quaternion
            obj = Quaternion()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # VRRig
    def Forward(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # VRRig
    def LeftHand(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = o + self._tab.Pos
            from .SimpleTransform import SimpleTransform
            obj = SimpleTransform()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # VRRig
    def RightHand(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = o + self._tab.Pos
            from .SimpleTransform import SimpleTransform
            obj = SimpleTransform()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # VRRig
    def Head(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = o + self._tab.Pos
            from .SimpleTransform import SimpleTransform
            obj = SimpleTransform()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def VRRigStart(builder): builder.StartObject(6)
def VRRigAddPosition(builder, position): builder.PrependStructSlot(0, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(position), 0)
def VRRigAddRotation(builder, rotation): builder.PrependStructSlot(1, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(rotation), 0)
def VRRigAddForward(builder, forward): builder.PrependStructSlot(2, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(forward), 0)
def VRRigAddLeftHand(builder, leftHand): builder.PrependStructSlot(3, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(leftHand), 0)
def VRRigAddRightHand(builder, rightHand): builder.PrependStructSlot(4, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(rightHand), 0)
def VRRigAddHead(builder, head): builder.PrependStructSlot(5, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(head), 0)
def VRRigEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class FlexParticleData(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsFlexParticleData(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = FlexParticleData()
        x.Init(buf, n + offset)
        return x

    # FlexParticleData
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # FlexParticleData
    def Id(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # FlexParticleData
    def Particles(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(tdw.flatbuffers.number_types.Uint8Flags, a + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # FlexParticleData
    def ParticlesAsNumpy(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(tdw.flatbuffers.number_types.Uint8Flags, o)
        return 0

    # FlexParticleData
    def ParticlesLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # FlexParticleData
    def Velocities(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(tdw.flatbuffers.number_types.Uint8Flags, a + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # FlexParticleData
    def VelocitiesAsNumpy(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(tdw.flatbuffers.number_types.Uint8Flags, o)
        return 0

    # FlexParticleData
    def VelocitiesLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def FlexParticleDataStart(builder): builder.StartObject(3)
def FlexParticleDataAddId(builder, id): builder.PrependInt32Slot(0, id, 0)
def FlexParticleDataAddParticles(builder, particles): builder.PrependUOffsetTRelativeSlot(1, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(particles), 0)
def FlexParticleDataStartParticlesVector(builder, numElems): return builder.StartVector(1, numElems, 1)
def FlexParticleDataAddVelocities(builder, velocities): builder.PrependUOffsetTRelativeSlot(2, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(velocities), 0)
def FlexParticleDataStartVelocitiesVector(builder, numElems): return builder.StartVector(1, numElems, 1)
def FlexParticleDataEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class AvatarChildrenNames(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsAvatarChildrenNames(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = AvatarChildrenNames()
        x.Init(buf, n + offset)
        return x

    # AvatarChildrenNames
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # AvatarChildrenNames
    def Id(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # AvatarChildrenNames
    def ChildIds(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, a + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # AvatarChildrenNames
    def ChildIdsAsNumpy(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(tdw.flatbuffers.number_types.Int32Flags, o)
        return 0

    # AvatarChildrenNames
    def ChildIdsLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # AvatarChildrenNames
    def ChildNames(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # AvatarChildrenNames
    def ChildNamesLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def AvatarChildrenNamesStart(builder): builder.StartObject(3)
def AvatarChildrenNamesAddId(builder, id): builder.PrependUOffsetTRelativeSlot(0, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(id), 0)
def AvatarChildrenNamesAddChildIds(builder, childIds): builder.PrependUOffsetTRelativeSlot(1, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(childIds), 0)
def AvatarChildrenNamesStartChildIdsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def AvatarChildrenNamesAddChildNames(builder, childNames): builder.PrependUOffsetTRelativeSlot(2, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(childNames), 0)
def AvatarChildrenNamesStartChildNamesVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def AvatarChildrenNamesEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class Collision(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsCollision(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = Collision()
        x.Init(buf, n + offset)
        return x

    # Collision
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # Collision
    def ColliderId(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Collision
    def CollideeId(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Collision
    def RelativeVelocity(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Collision
    def State(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 1

    # Collision
    def Contacts(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Vector(o)
            x += tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j) * 24
            from .ContactPoint import ContactPoint
            obj = ContactPoint()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Collision
    def ContactsLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def CollisionStart(builder): builder.StartObject(5)
def CollisionAddColliderId(builder, colliderId): builder.PrependInt32Slot(0, colliderId, 0)
def CollisionAddCollideeId(builder, collideeId): builder.PrependInt32Slot(1, collideeId, 0)
def CollisionAddRelativeVelocity(builder, relativeVelocity): builder.PrependStructSlot(2, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(relativeVelocity), 0)
def CollisionAddState(builder, state): builder.PrependUint8Slot(3, state, 1)
def CollisionAddContacts(builder, contacts): builder.PrependUOffsetTRelativeSlot(4, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(contacts), 0)
def CollisionStartContactsVector(builder, numElems): return builder.StartVector(24, numElems, 4)
def CollisionEnd(builder): return builder.EndObject()
